/**
 ** scoutExecTaskCodels.c
 **
 ** Codels appeles par la tache scoutExecTask
 **
 ** Auteur: 
 ** Date: 
 **
 **/

#include <portLib.h>

#include <stdio.h>
#include <string.h>
#include "h2mathLib.h"
#include "nomadic/Nclient.h"

#include "server/scoutHeader.h"
#include "scoutConst.h"


#include "server/scoutHeader.h"
#include "scoutConst.h"
#include "scoutServoCntrlLaws.h"
#include "scoutCmdLibProto.h"


/*------------------ Variables globales ------------------*/

static double scoutExecTaskPeriod;

/*----------------------------------------------------------------------
 *                  PROTOTYPE DES FONCTIONS LOCALES
 */ 

static BOOL alarmStatus(SCOUT_ALARMS_TYPE alarm);
static STATUS scoutAlarmVerify (int scoutImportantDriftFlag,
				int *bilan);
static STATUS scoutVelocitySet (double v, double w);
static STATUS scoutVelocityControl (SCOUT_SPEED_REF_STR *refId,
				    SCOUT_POS_STR *posId);

static void scoutSpeedGlobToWheels(double v, double w, double *vr, double *vl);
static void scoutCntrlRobotMotion(double prevV, double prevW,
				  int *scoutExecTaskNoMotionCounter);


/*--  PERMANENTE ACTIVITY   -------------------------------------------
 *
 *  odoAndAsserv - Calcul position et asservissement
 *
 *  Parametres: 
 *
 *  Retourne: OK ou ERROR
 */

STATUS servoCntrl(int *bilan)
{
  static BOOL scoutImportantDriftFlag=FALSE;

  /* 
   * Alarms  states
   */
  if (scoutAlarmVerify (scoutImportantDriftFlag, bilan) != OK)
    return(ERROR);
  

  switch (SDI_F->motionType) {

  /*
   *  No motion or Nomadic motion 
   */
  case SCOUT_NO_MOTION:
  case SCOUT_NOMADIC_MOTION:
    scoutSetRefToCurrentPos();
    SDI_F->speedMotorRef.v = 0.0;
    SDI_F->speedMotorRef.w = 0.0;
    scoutImportantDriftFlag = FALSE;
    return OK;

  /* 
   * Servo control of the robot
   */
    
    /* Position control */
  case SCOUT_SERVO_POS:
    scoutImportantDriftFlag = scoutPositionControl (&SDI_F->speedMotorRef.v,
						    &SDI_F->speedMotorRef.w);
    if (scoutVelocitySet (SDI_F->speedMotorRef.v, SDI_F->speedMotorRef.w) != OK)
      return(ERROR);
    return OK;

    /* Speed control */
  case SCOUT_SERVO_SPEED:
    scoutImportantDriftFlag = scoutSpeedControl (&SDI_F->speedRef, 
						 &SDI_F->speedRef);

    /* En prevision du retour d'asservissement en position 
       on met a jour la consigne en position avec la position courante */
    SDI_F->ref.x = SDI_F->odo.robot.x;
    SDI_F->ref.y = SDI_F->odo.robot.y;
    SDI_F->ref.theta = SDI_F->odo.robot.theta;

    /* Imposer au robot les vitesses lineaire et angulaire de consigne */
    SDI_F->speedMotorRef.v = SDI_F->speedRef.v;
    SDI_F->speedMotorRef.w = SDI_F->speedRef.w;
    if (scoutVelocityControl (&SDI_F->speedRef, &SDI_F->odo.robot) == ERROR)
      return(ERROR);
    return OK;

    /* Configuration control */
  case SCOUT_SERVO_CONFIG:
    scoutImportantDriftFlag = scoutConfigurationControl (&SDI_F->speedMotorRef.v, 
							 &SDI_F->speedMotorRef.w);
    if (scoutVelocitySet (SDI_F->speedMotorRef.v, SDI_F->speedMotorRef.w) != OK)
      return(ERROR);
    return OK;

  }
  
  return(OK);
}

/*----------------------------------------------------------------------
 *
 *  Fonction d'initialisation des taches d'executions
 *
 *  Parametres: Structure de donnees interne
 *
 *  Retourne: OK ou ERROR
 */

STATUS initOdoAndServo(int *bilan)

{
  SCOUT_CMD_PARAM_STR *cmdId;       /* Id parametres boucle asservissement */
  SCOUT_REF_STR *refId;             /* Id donnees de consigne 
				       boucle asservissement */
  SCOUT_ODO_STR *odoId;             /* Id donnees odometriques (roues odo) */
  
  /* Pas d'alarme */
/*   memset(SDI_F->alarms, 0, sizeof(SCOUT_ALARMS)); */

  /*Obtenir l'id des parametres cinematiques boucle d'asservissement */
  cmdId = &scoutDataStrId->cmd;

  /* Init les valeurs par defaut des parametres boucle asservissement */
  cmdId->distPoint = SCOUT_DIST_TO_CONTROLED_POINT;
  cmdId->KpLongit = SCOUT_KP_LONGIT; 
  cmdId->KiLongit = SCOUT_KI_LONGIT;
  cmdId->KpTransv = SCOUT_KP_TRANSV;
  cmdId->KiTransv = SCOUT_KI_TRANSV;
  cmdId->KpAng    = SCOUT_KP_ANG;
  cmdId->KiAng    = SCOUT_KI_ANG;
  cmdId->longitErrorMax = SCOUT_LONGIT_ERROR_MAX;
  cmdId->longitMaxErrSum = SCOUT_LONGIT_MAX_ERR_SUM;
  cmdId->transvErrorMax = SCOUT_TRANSV_ERROR_MAX;
  cmdId->transvMaxErrSum = SCOUT_TRANSV_MAX_ERR_SUM;
  cmdId->angErrorMax = SCOUT_ANG_ERROR_MAX;
  cmdId->angMaxErrSum = SCOUT_ANG_MAX_ERR_SUM;

  /* Initialiser le controle des roues */
  scoutExecTaskPeriod = EXEC_TASK_PERIOD(SCOUT_EXECTASK_NUM);
  if (scoutExecTaskPeriod <= 0) {
    printf ("scoutExecTaskPeriod: %f !!!\n", scoutExecTaskPeriod);
    return(ERROR);
  }

  /* Obtenir l'id de la structure de donnees de consigne */
  refId = &scoutDataStrId->ref;

  SDI_F->scoutWaitRefMax = 62;

  /* Initialiser les valeurs de consignes */
  refId->backFlag = SCOUT_FORWARD_MOTION;
  refId->numRef = 0;
  refId->x = cmdId->distPoint;
  refId->y = 0.;
  refId->theta = 0.;
  refId->v = 0.;
  refId->w = 0.;
  refId->vmax = SCOUT_VMAX_ROBOT;
  refId->wmax = SCOUT_WMAX_ROBOT;
  refId->linAccelMax = SCOUT_AMAX_LIN_ROBOT;
  refId->angAccelMax = SCOUT_AMAX_ANG_ROBOT;

  SDI_F->speedRef.numRef = 0;
  SDI_F->speedRef.v = 0.0;
  SDI_F->speedRef.w = 0.0;
  SDI_F->speedRef.vmax = SCOUT_VMAX_ROBOT;
  SDI_F->speedRef.wmax = SCOUT_WMAX_ROBOT;
  SDI_F->speedRef.linAccelMax = SCOUT_AMAX_LIN_ROBOT;
  SDI_F->speedRef.angAccelMax = SCOUT_AMAX_ANG_ROBOT;

  SDI_F->speedMotorRef.v = 0.0;
  SDI_F->speedMotorRef.w = 0.0;

  /* Obtenir l'id de l'odometrie (roues odo) */
  odoId = &scoutDataStrId->odo;

  /* Initialiser la position du robot (point controle a l'origine) */
  odoId->robot.x = cmdId->distPoint;
  odoId->robot.y = 0;
  odoId->robot.xC = 0.;
  odoId->robot.yC = 0;
  odoId->robot.theta = 0;
  odoId->robot.v = 0;
  odoId->robot.w = 0;

  /* Init de l'erreur sur l'odo probabiliste */
  memset(odoId->odoError.var, 0, sizeof(SCOUT_ODO_VAR_STR));
  odoId->odoError.var[0] = SCOUT_INIT_VAR_X;
  odoId->odoError.var[2] = SCOUT_INIT_VAR_Y;
  odoId->odoError.var[5] = SCOUT_INIT_VAR_THETA;

  SDI_F->motionType = SCOUT_NO_MOTION;
  SDI_F->trackStr.scoutCmdTrackMode = SCOUT_NO_TRACKING;
  return(OK);
}



/*------ FONCTIONS AUXILIAIRES de odoAndAsserv ---------------------------*/



/*----------------------------------------------------------------------
 *
 *  scoutVelocitySet  -  Imposer les vitesses aux roues du robot
 *
 *  Description:
 *  Cette fonction calcule, a partir des vitesses angulaire et lineaire du 
 *  robot (fournies en entree) les vitesses a imposer aux roues motrices, qui
 *  sont ensuite actionnees avec ces vitesses
 *
 *  Retourne: OK ou ERROR
 */

static STATUS scoutVelocitySet (double v,  /* Vitesse lineaire de commande */
				double w)  /* Vitesse angulaire de commande */

{
  double vr, vl;

  /* vitesses lineaires roues droite et gauche */
  vr = v + SCOUT_DIAMETER * w /2.0;
  vl = v - SCOUT_DIAMETER * w /2.0;

  /* Vitesse lineaire pour chaque roue */
  vm((int)M_TO_DECI_IN(vr), (int)M_TO_DECI_IN(vl),0);

  return OK;
}



/*----------------------------------------------------------------------
 *
 *  scoutVelocityControl  -  Imposer les vitesses aux roues du robot
 *
 *  Description:
 *  Cette fonction calcule, a partir des vitesses angulaire et lineaire du 
 *  robot (fournies en entree) les vitesses a imposer aux roues motrices, qui
 *  sont ensuite actionnees avec ces vitesses
 *
 *  Retourne: OK ou ERROR
 */


static STATUS 
scoutVelocityControl (SCOUT_SPEED_REF_STR *refId,
		      SCOUT_POS_STR *posId)

{
  double vLeftMax;          /* Vitesse max de la roue gauche */
  double vRightMax;         /* Vitesse max de la roue droite */
  double accelLeftMax;      /* Accel max roue gauche */
  double accelRightMax;     /* Accel max roue droite */
  static double vLeftPrev=0.0;
  static double vRightPrev=0.0;

  double vLeftRef;          /* Vitesse consigne roue gauche */
  double vRightRef;         /* Vitesse consigne roue droite */
  double vLeft;             /* Vitesse actuelle de la roue gauche */
  double vRight;            /* Vitesse actuelle de la roue droite */

  /* Vitesses actuelles 
     (a partir de la vitesse instantannee (v,w) du robot) */
  scoutSpeedGlobToWheels(posId->v, posId->w, &vRight, &vLeft);

  /* Vitesses consigne  */
  scoutSpeedGlobToWheels(refId->v, refId->w, &vRightRef, &vLeftRef);

  /* Vitesse max (elle est sur la roue droite) */
  scoutSpeedGlobToWheels(refId->vmax, refId->wmax, &vRightMax, &vLeftMax);
  vLeftMax = vRightMax;

  /* Accel max (elle est sur la roue droite) */
  scoutSpeedGlobToWheels(refId->linAccelMax, refId->angAccelMax, 
			&accelRightMax, &accelLeftMax);
  accelLeftMax = accelRightMax;

  /* Vitesse lineaire pour chaque roue */
  vm((int)M_TO_DECI_IN(vRightRef), 
     (int)M_TO_DECI_IN(vLeftRef), 0);
  
  vRightPrev = vRight;
  vLeftPrev = vLeft;
  return OK;
}

/*----------------------------------------------------------------------
 *
 *  scoutSpeedGlobToWheels - Conversion de vitesses lineaire et angulaire
 *                          du robot en des vitesses angulaires des roues
 *
 *  Retourne: rien
 */
static void
scoutSpeedGlobToWheels(double v, double w, double *vr, double *vl)
{
  *vr = v + SCOUT_DIAMETER * w /2.0;
  *vl = v - SCOUT_DIAMETER * w /2.0;
}


/*----------------------------------------------------------------------
 *
 *  scoutCntrlRobotMotion
 *
 *  Description: 
 *  Cette fonction permet de verifier que le robot bouge
 *  quand on lui envoie des consignes
 *
 *  Retourne: OK ou ERROR
 */
static void scoutCntrlRobotMotion(double vRef, double wRef,
				  int *scoutExecTaskNoMotionCounter)
{

  /* Verifier que le robot a bouge */
  if ((fabs(vRef) > 0.01 || fabs(wRef) > 0.01) 
      && (fabs(SDI_F->odo.robot.v) < 0.0001 && 
	  fabs(SDI_F->odo.robot.w) < 0.0001)) {
    (*scoutExecTaskNoMotionCounter)++;
    logMsg("(v, w) cmd: %.3f %.3f\n", vRef, wRef);
    logMsg(" -> %.3f %.3f (%d)\n", SDI_F->odo.robot.v, 
	   SDI_F->odo.robot.w, *scoutExecTaskNoMotionCounter);
  }
  else 
    *scoutExecTaskNoMotionCounter = 0;
}


/*----------------------------------------------------------------------
 *
 *  scoutAlarmVerify - Verifier l'etat des alarmes
 *
 *  Description: 
 *  Cette fonction permet de verifier l'etat des signaux d'alarme et de
 *  reagir en fonction. (surcouple, coup de poing, choc arceau, etc)
 *
 *  Retourne: OK ou ERROR
 */

static STATUS scoutAlarmVerify (int scoutImportantDriftFlag,
				int *bilan)

{
  BOOL alarmOn;             /* Etat de l'alarme */
  BOOL prevAlarm;           /* Etat precedent de l'alarme */
  BOOL alarmManual;         /* Mode manuel */
  int switchAllOff = FALSE;    /* Flag pour tout arreter si alarme */
  int switchAllOn = FALSE;     /* Flag pour tout demarrer si plus d'alarme */
  static int scoutRefNotUpdated=FALSE;
  static int prevNumRef=2;
  int numRef;
  static int scoutExecTaskNoMotionCounter=0;


  /*
   * Batteries
   */
  prevAlarm = SDI_F->alarms.lowBatteryFlag;
  alarmOn = alarmStatus (SCOUT_ALARM_BATTERY);
  if (alarmOn)  *bilan = S_scout_LOW_BATTERY;
  /* Changement etat */
  if (prevAlarm != alarmOn) {
    /* Batteries dechargees */
    if (alarmOn) {
      SDI_F->alarms.nbAlarms++;
      SDI_F->alarms.lowBatteryFlag = TRUE;
      logMsg("S_scout_LOW_LEVEL_BATTERY");
    }
    else {
      SDI_F->alarms.nbAlarms--;
      SDI_F->alarms.lowBatteryFlag = FALSE;
    }
  }

  /* 
   * Arceau 
   */
  prevAlarm = SDI_F->alarms.chocFlag;
  alarmOn = alarmStatus (SCOUT_ALARM_CHOC);
  if (alarmOn) *bilan = S_scout_CHOC_BUMPER;
  /* Changement etat */
  if (prevAlarm != alarmOn) {
    /* Chock arceau */
    if (alarmOn) {
      SDI_F->alarms.nbAlarms++;
      switchAllOff = TRUE;
      SDI_F->alarms.chocFlag = TRUE;
      logMsg("S_scout_CHOC_BUMPER\n");
    }
    else {
      SDI_F->alarms.nbAlarms--;
      switchAllOn = TRUE;
      SDI_F->alarms.chocFlag = FALSE;
    }
  }

  /* 
   * Arret d'urgence 
   */
  prevAlarm = SDI_F->alarms.estopDownFlag;
  alarmOn = alarmStatus (SCOUT_ALARM_ESTOP);
  if (alarmOn) 
    *bilan =  S_scout_ESTOP_DOWN;
  if (prevAlarm != alarmOn) {
    if (alarmOn) {
      SDI_F->alarms.nbAlarms++;
      switchAllOff = TRUE;
      SDI_F->alarms.estopDownFlag = TRUE;
      logMsg("S_scout_ESTOP_DOWN\n");
    }
    else{
      SDI_F->alarms.nbAlarms--;
      switchAllOn = TRUE;
      SDI_F->alarms.estopDownFlag = FALSE;
    }
  }
  
  /* 
   * Switch auto <-> manuel 
   */
  prevAlarm = SDI_F->alarms.joystickInUseFlag;
  alarmOn = alarmStatus (SCOUT_ALARM_JOYSTICK);
  if (alarmOn) 
    *bilan =  S_scout_JOYSTICK_IN_USE;
  alarmManual = alarmOn;
  if (prevAlarm != alarmOn) {
    if (alarmOn) {
      SDI_F->alarms.nbAlarms++;
      switchAllOff = TRUE;
      SDI_F->alarms.joystickInUseFlag  = TRUE;
      logMsg("S_scout_JOYSTICK_IN_USE\n");
    }
    else{
      SDI_F->alarms.nbAlarms--;
      switchAllOn = TRUE;
      SDI_F->alarms.joystickInUseFlag  = FALSE;
    }
  }

  /* 
   * No motion
   */
  prevAlarm = SDI_F->alarms.noMotionFlag;
  scoutCntrlRobotMotion(SDI_F->speedMotorRef.v,
			SDI_F->speedMotorRef.w,
			&scoutExecTaskNoMotionCounter);
  alarmOn = (scoutExecTaskNoMotionCounter) > 5 ? TRUE : FALSE;
  if (alarmOn) {
    *bilan =  S_scout_NO_MOTION;
    scoutExecTaskNoMotionCounter = 0;
  }
  if (prevAlarm != alarmOn) {
    if (alarmOn) {
      SDI_F->alarms.nbAlarms++;
      switchAllOff = TRUE;
      SDI_F->alarms.noMotionFlag = TRUE;
      logMsg("S_scout_NO_MOTION\n");
    }
    else{
      SDI_F->alarms.nbAlarms--;
      switchAllOn = TRUE;
      SDI_F->alarms.noMotionFlag = FALSE;
    }
  }

  /* 
   * Important drift during servo-control
   */
  prevAlarm = SDI_F->alarms.importantDriftFlag;
  alarmOn = scoutImportantDriftFlag == 1 ? TRUE : FALSE;
  if (alarmOn) 
    *bilan =  S_scout_IMPORTANT_DRIFT;
  if (prevAlarm != alarmOn) {
    if (alarmOn) {
      SDI_F->alarms.nbAlarms++;
      switchAllOff = TRUE;
      SDI_F->alarms.importantDriftFlag = TRUE;
      logMsg("S_scout_IMPORTANT_DRIFT\n");
    }
    else{
      SDI_F->alarms.nbAlarms--;
      switchAllOn = TRUE;
      SDI_F->alarms.importantDriftFlag = FALSE;
    }
  }


  /* 
   * Tracking reference not updated
   *
   * scoutPumpTask signal la mise a jour de la consigne en faisant 
   *    varier la valeur de  scoutDataStrId->ref.numRef. 
   *    si ca ne switch plus -> alarm (et numRef mis a  -1)
   * Particularite: 
   *    Tant que cette alarm n'a pas ete detectee et annoncee au client
   *    on la maintient. 
   * si scoutPumpTask se reactive, il le signal en remettant numRef a 0
   *    ce qui desactive l'alarm 
   */
  prevAlarm = SDI_F->alarms.refNotUpdatedFlag;
  /* Etat courant alarm */
  if (scoutDataStrId->trackStr.scoutCmdTrackMode) {
    numRef = SDI_F->motionType == SCOUT_SERVO_SPEED ?
	     scoutDataStrId->speedRef.numRef : scoutDataStrId->ref.numRef;
    if (numRef == prevNumRef) {
      scoutRefNotUpdated++;
      /*      printf ("scoutRefNotUpdated %d (%d)", scoutRefNotUpdated, prevNumRef);*/
    }
    else {
      prevNumRef = numRef;
      scoutRefNotUpdated = FALSE;
    }
  }
  alarmOn = scoutRefNotUpdated > SDI_F->scoutWaitRefMax ? TRUE : FALSE;
  if (alarmOn) {
    /* Nouvelle alarm */
    if (prevAlarm != alarmOn) {
      SDI_F->alarms.nbAlarms++;
      scoutDataStrId->ref.numRef = -1;
      scoutDataStrId->speedRef.numRef = -1;
      switchAllOff = TRUE;
      SDI_F->alarms.refNotUpdatedFlag = TRUE;
      *bilan =  S_scout_REF_NOT_UPDATED;
      logMsg("S_scout_REF_NOT_UPDATED\n");
    }

    /* Maintient etat alarm tant que non detectee */
    else if (SDI_F->ref.numRef== -1 || SDI_F->speedRef.numRef== -1)
      *bilan =  S_scout_REF_NOT_UPDATED;

    /* Alarm detectee: on desactive l'alarm  */
    else {
      SDI_F->alarms.nbAlarms--;
      scoutRefNotUpdated = FALSE;
      SDI_F->alarms.refNotUpdatedFlag = FALSE;
      switchAllOn = TRUE;
    }
  }


  /* 
   * Si changement d'etat et alarme activee
   */ 
  /* Arret tracking et moteurs */
  if (switchAllOff) {
    SDI_F->motionType = SCOUT_NO_MOTION;
    scoutDataStrId->trackStr.scoutCmdTrackMode = SCOUT_NO_TRACKING;
    scoutSetRefToCurrentPos();
    scoutStopExec(bilan);
  }

  /* Remise en route moteurs */
  else if (switchAllOn) {
    scoutSetRefToCurrentPos();
  }

  /* Si on est en manuel */
  else if (alarmManual) {
    scoutSetRefToCurrentPos();
  }
  return (OK);
}

/*----------------------------------------------------------------------
 * 
 *   scoutSetRefToPos - set ref to current position
 *
 */
void scoutSetRefToCurrentPos (void) 
{
  SCOUT_REF_STR *refId = &scoutDataStrId->ref;
  SCOUT_POS_STR *robotId = &(SDI_F->odo.robot);

  /* Consignes = odometrie */
  refId->x = robotId->x;
  refId->y = robotId->y;
  refId->theta = robotId->theta;
  refId->v = 0.;
  refId->w = 0.;
  refId->vmax = SCOUT_VMAX_ROBOT;
  refId->wmax = SCOUT_WMAX_ROBOT;
  refId->linAccelMax = SCOUT_AMAX_LIN_ROBOT;
  refId->angAccelMax = SCOUT_AMAX_ANG_ROBOT;

  SDI_F->speedRef.v = 0.0;
  SDI_F->speedRef.w = 0.0;
}    

/*----------------------------------------------------------------------
 *
 * alarmStatus  - updates the alarms status
 *
 *
 *
 */
static BOOL alarmStatus(SCOUT_ALARMS_TYPE alarm)
{
  int i;

  switch(alarm) {

  case SCOUT_ALARM_CHOC:
    for (i=0; i<SCOUT_NB_BUMPERS; i++)
      if (SDI_F->bumpersState.bumper[i])
	return TRUE;
    return FALSE;

  case SCOUT_ALARM_BATTERY:
    if(SDI_F->batteriesLevel.charge == SCOUT_BATTERY_LOW)
      return TRUE;
    return FALSE;
      
  case SCOUT_ALARM_JOYSTICK:
    /* On ne sait pas detecter... */
    return FALSE;

  case SCOUT_ALARM_ESTOP:
    /* On ne sait pas detecter... */
    return FALSE;

  case  SCOUT_ALARM_MOTION_ERROR:
    return FALSE;

  case  SCOUT_ALARM_NO_MOTION:
    return FALSE;

  case  SCOUT_ALARM_DRIFT:
    return FALSE;

  case  SCOUT_ALARM_REF_NOT_UPDATE:
    return FALSE;

  default:
    return FALSE;
  }
}
