/**
 ** scoutExecTaskCodels.c
 **
 ** Codels appeles par la tache scoutExecTask
 **
 ** Auteur: 
 ** Date: 
 **
 **/

#include <math.h>
#include <stdio.h>
#include <string.h>

#include <portLib.h>
#include <h2mathLib.h>

#include "nomadic/Nclient.h"

#include "server/scoutHeader.h"
#include "scoutConst.h"
#include "scoutCmdLibProto.h"
#include "scoutVarianceProto.h"
#include "scoutSonarsProto.h"

static STATUS odoAcquisition (SCOUT_POS_STR *posId);
static void scoutGetBatteriesLevel(int printFlag);

/*------------------------------------------------------------------------
 *
 * scoutExecTaskInit  --  Codel d'initialisation (SDI/f, ...)
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS scoutExecTaskInit(int *bilan)
{
  int i;
  int sonarsOrder[SCOUT_NB_SONARS] = SCOUT_DEFAULT_SONARS_FIRING_ORDER;


  *bilan = OK;

  /* Init odo and servo */
  if (initOdoAndServo(bilan) == ERROR) {
    return ERROR;
  }
  
  /* Sonars */
  SDI_F->sonarsFiring.fireRate = SCOUT_DEFAULT_SONARS_FIRING_RATE;
  for (i=0; i<SCOUT_NB_SONARS; i++) {
    SDI_F->sonarsFiring.firedSonars[i] = sonarsOrder[i];
  }
  scoutComputeSonarsCoord(bilan);
  scoutComputeBumpersCoord(bilan);

  /* Connection */
  SDI_F->connectParam.connectMode = SCOUT_DEFAULT_CONNECT_MODE;
  strcpy(SDI_F->connectParam.host, SCOUT_DEFAULT_CONNECT_HOST);
  SDI_F->connectParam.port = SCOUT_DEFAULT_CONNECT_PORT;
  SDI_F->connectParam.robotId  = SCOUT_DEFAULT_CONNECT_ROBOT_ID;
  SDI_F->connectStatus = SCOUT_NOT_CONNECTED;

  scoutConnectRobotExec(&SDI_F->connectParam, bilan);

  return OK;
}

/*------------------------------------------------------------------------
 *
 * scoutExecTaskEnd  --  Codel de terminaison
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS scoutExecTaskEnd(void)
{
  int bilan;
  /* Test connection */
  if (SDI_F->connectStatus == SCOUT_NOT_CONNECTED)
    return OK;
  /* Disconnect */
  scoutDisconnectRobotExec(&bilan);
  return OK;
}

/*------------------------------------------------------------------------
 *
 * scoutExecTaskPerm  --  codel de l'activite' permanente
 *
 * Description: 
 * 
 * Bilans:      OK 
 *              S_scout_UNKNOWN_AXIS_STATUS
 *              S_scout_INVALID_ARGUMENT
 *              S_scout_JOYSTICK_IN_USE
 *              S_scout_ESTOP_DOWN
 *              S_scout_MOTION_ERROR
 *              S_scout_CHOC_BUMPER
 *              S_scout_LOW_BATTERY
 * 
 * Retourne:    OK ou ERROR
 */

STATUS scoutExecTaskPerm(int *bilan)
{

  /* Test connection */
  if (SDI_F->connectStatus == SCOUT_NOT_CONNECTED)
    return OK;

  /* Read all sensors */
  if (!gs()) {
    printf("scoutExecTaskPerm: cannot read sensors. Disconnect server.\n");
    SDI_F->connectStatus = SCOUT_NOT_CONNECTED;
  }

  /* Batteries */
  scoutGetBatteriesLevel(FALSE);

  /* Sonars to poster */
  if (scoutComputeSonarsEcho(bilan) != OK) {
    return ERROR;
  }

  /* Bumbers to poster */
  if (scoutComputeBumpersEcho(bilan) != OK) {
    return ERROR;
  }

  /* Odo acquisition */
  if (odoAcquisition(&SDI_F->odo.robot) == ERROR) {
    printf("scoutExecTaskPerm: odoAcquisition failed\n");
    return ERROR;
  }

  /* Servo control */
  if (servoCntrl(bilan) == ERROR) {
    printf("scoutExecTaskPerm: servoCntrl failed\n");
    return ERROR;
  }

  return OK;
}

/*------------------------------------------------------------------------
 * ConnectRobot
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_scout_ROBOT_NOT_FOUND
 *              S_scout_CONNECTION_FAILED
 *              S_scout_OUT_OF_MEMORY
 *              S_scout_UNKNOWN_ERROR
 */

/* scoutConnectRobotExec  -  codel EXEC de ConnectRobot
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT scoutConnectRobotExec (SCOUT_CONNECT_PARAM *connectParam,
				      int *bilan)
{
  int status;

  /* Test connection */
  if (SDI_F->connectStatus != SCOUT_NOT_CONNECTED) {
    printf ("scoutConnectRobotExec: already connected\n");
  }

  /* Try to connect */
  else if (connectParam->connectMode != SCOUT_NOT_CONNECTED) {

    /* Parameters for Nclient */
    SERV_TCP_PORT = connectParam->port;
    strcpy(SERVER_MACHINE_NAME, connectParam->host);

    /* Connection */
    status = connect_robot(connectParam->robotId, 
			   MODEL_SCOUT2, connectParam->host, 38400);
    
    /* Set connection status */
    if (status) {
      printf ("scoutExecTaskInit: Connected to %s\n", 
	      connectParam->host);
      SDI_F->connectStatus = connectParam->connectMode;
    }
    else {
      printf ("scoutExecTaskInit: CAN NOT connect %s\n", 
	      connectParam->host);
      *bilan = S_scout_CONNECTION_FAILED;
    }
  } /* New connection */

  /* Real <-> Simul */
  SDI_F->nclientSimulSpeedFactor = 1;
  if(connectParam->connectMode == SCOUT_NCLIENT_CONNECTION)
    real_robot();
  if(connectParam->connectMode == SCOUT_NCLIENTSIMUL_CONNECTION) {
    simulated_robot();
    SDI_F->nclientSimulSpeedFactor = State[STATE_SIM_SPEED]/10;
    printf ("Speed simul factor = %g\n", SDI_F->nclientSimulSpeedFactor);
  }

  /* no connection: stop there */
  if (*bilan != OK || SDI_F->connectStatus == SCOUT_NOT_CONNECTED) {
    printf ("scoutConnectRobotExec: no connection\n");
    return ETHER;
  }

  /* Init commande */
  conf_tm(5);

  /* Init position */
  dp(SDI_F->odo.robot.xC,SDI_F->odo.robot.yC);
  da(SDI_F->odo.robot.theta,0);


  /* Read batteries */
  scoutGetBatteriesLevel(TRUE);

  return ETHER;
}

/*------------------------------------------------------------------------
 * DisconnectRobot
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_scout_UNINITIALIZED
 *              S_scout_ROBOT_NOT_FOUND
 *              S_scout_CONNECTION_FAILED
 */

/* scoutDisconnectRobotExec  -  codel EXEC de DisconnectRobot
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT scoutDisconnectRobotExec (int *bilan)
{
  disconnect_robot(SDI_F->connectParam.robotId);
  return ETHER;
}

/*------------------------------------------------------------------------
 * Turn
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_scout_UNINITIALIZED
 *              S_scout_NO_MOTION
 */


/* scoutTurnStart  -  codel EXEC de Turn
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT scoutTurnStart(SCOUT_TURN *turn, int *bilan)
{
  double dr, dl;  /* variation curviligne roues droite et gauche */
  double vr, vl;  /* vitesses lineaires roues droite et gauche */
  double v, w;    /* vitesse lineaire et angulaire du robot */

  /* variation curviligne roues droite et gauche */
  dr = (2 * turn->dl + SCOUT_DIAMETER * turn->dth)/2.0;
  dl = (2 * turn->dl - SCOUT_DIAMETER * turn->dth)/2.0;

  /* vitesse lineaire et angulaire du robot */
  if (fabs(turn->dl) > 0.0001) {
    v = turn->speedPercent * SCOUT_VMAX_ROBOT/100.;
    w = v * turn->dth / turn->dl;
  }
  else if (fabs(turn->dth) > 0.0001) {
    w = turn->speedPercent * SCOUT_WMAX_ROBOT/100;
    v = w * turn->dl / turn->dth;
  }

  /* vitesses lineaires roues droite et gauche */
  vr = v + SCOUT_DIAMETER * w /2.0;
  vl = v - SCOUT_DIAMETER * w /2.0;

  printf ("Turn: dr=%f m/s dl=%frd/s\n", dr,dl);
  printf ("Turn: vr=%f m/s vl=%frd/s\n", vr,vl);

  SDI_F->motionType = SCOUT_NOMADIC_MOTION;

  /* Acceleration */
  ac(200,200,0);

  /* Vitesse lineaire pour chaque roue */
  sp((int)(SDI_F->nclientSimulSpeedFactor*M_TO_DECI_IN(vr)), 
     (int)(SDI_F->nclientSimulSpeedFactor*M_TO_DECI_IN(vl)),
     0);

  /* Longueur de parcours de chaque roue */
  pr((int)M_TO_DECI_IN(dr), (int)M_TO_DECI_IN(dl),0);

  return EXEC;
}

/* scoutTurnExec  -  codel EXEC de Turn
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT scoutTurnExec (SCOUT_TURN *turn, int *bilan)
{
  
  if ( ws(1,1,0,1)) {
    SDI_F->motionType = SCOUT_NO_MOTION;
    return ETHER;
  }
  return EXEC;
}

/*------------------------------------------------------------------------
 * Stop
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* scoutStopExec  -  codel EXEC de Stop
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT scoutStopExec (int *bilan)
{
/*    sp(0,0,0); */
  vm(0,0,0);
  scoutSetRefToCurrentPos();
  SDI_F->motionType = SCOUT_NO_MOTION;
  SDI_F->trackStr.scoutCmdTrackMode = SCOUT_NO_TRACKING;
  return ETHER;
}

/*------------------------------------------------------------------------
 * SetPos
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_scout_TRACKING_ON
 */

/* scoutSetPosExec  -  codel EXEC de SetPos
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT scoutSetPosExec (SCOUT_SET_POS_STR *setPos, int *bilan)
{
  SCOUT_REF_STR *refId = &(SDI_F->ref);
  double xRef, yRef;             /* Consigne pour le point en avant */
  double dist = SDI_F->cmd.distPoint;
  SCOUT_POS_STR *posId;
  
  posId = &SDI_F->odo.robot;
  
  /* Limiter l'angle de recalage */
  setPos->theta = angleLimit (setPos->theta);
  
  /* Calculer la consigne pour le point en avant */
  /* recalage du centre du robot */
  xRef = setPos->x + dist * cos (setPos->theta);
  yRef = setPos->y + dist * sin (setPos->theta);
  /* Recalage statique de la consigne : consigne=position recale'e */
  refId->x = xRef;
  refId->y = yRef;
  refId->theta = setPos->theta;
  
  refId->v = 0;
  refId->w = 0;
  
  /* Controle point en avant */
  refId->backFlag = SCOUT_FORWARD_MOTION;
  
  /* Arrete l'asservissement */
  scoutStopExec(bilan);
/*   if (scoutStopAsserv(setPos->x, setPos->y, setPos->theta, bilan) == ERROR) { */
/*     return ETHER; */
/*   } */

  /* Update odometry */
  if (!dp(M_TO_DECI_IN(setPos->x), M_TO_DECI_IN(setPos->y))) {
    printf("scoutSetPosExec: dp error\n");
    *bilan = S_scout_UNKNOWN_ERROR;
    return ETHER;
  }
  if (!da(RAD_TO_DECI_DEG(setPos->theta),0)){ 
    printf("scoutSetPosExec: da error\n");
    *bilan = S_scout_UNKNOWN_ERROR;
    return ETHER;
  }

  posId->xC = setPos->x;
  posId->yC = setPos->y;
  posId->theta = setPos->theta;
    
  return ETHER;
}

/*------------------------------------------------------------------------
 * GoTo
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_scout_BAD_REF_DATA
 *              S_scout_TRACKING_ON
 */

/* scoutGoToStart  -  codel START de GoTo
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT scoutGoToStart (SCOUT_REF_STR *ref, int *bilan)
{
  return ETHER;
}

/* scoutGoToExec  -  codel EXEC de GoTo
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT scoutGoToExec (SCOUT_REF_STR *ref, int *bilan)
{
  if ( ws(1,1,0,1)) {
/*      SDI_F->motionType = SCOUT_NO_MOTION; */
    return ETHER;
  }
  return EXEC;
}


/*------------------------------------------------------------------------
 * GoToSpeed
 *
 * Description: 
 *
 * Bilans:      EXEC 
 *              S_scout_BAD_REF_DATA
 *              S_scout_TRACKING_ON
 */

/* scoutGoToSpeedStart  -  codel START de GoToSpeed
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT scoutGoToSpeedStart (SCOUT_SPEED_REF_STR *speedRef, int *bilan)
{

  return ETHER;
}

/* scoutGoToSpeedExec  -  codel EXEC de GoToSpeed
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT scoutGoToSpeedExec (SCOUT_SPEED_REF_STR *speedRef, int *bilan)
{
  /* ... add your code here ... */
  return ETHER;
}





/*****************************************************************************
 *
 *  odoAcquisition  -  Acquisition des donnees odometriques
 *
 *  Description:
 *  Cette fonction permet de lire les codeurs des roues odometriques et de
 *  calculer la position, orientation, vitesses, etc actuelles du robot.
 *
 *  Retourne: OK ou ERROR
 */

static STATUS odoAcquisition (SCOUT_POS_STR *posId)

{
  int sens;
  double distPoint = SDI_F->cmd.distPoint;
  double ds, dt;
  double scoutExecTaskPeriod = EXEC_TASK_PERIOD(SCOUT_EXECTASK_NUM);

  /* Position */
  posId->xC = DECI_IN_TO_M(State[STATE_CONF_X]);
  posId->yC = DECI_IN_TO_M(State[STATE_CONF_Y]);
  posId->theta = DECI_DEG_TO_RAD(State[STATE_CONF_STEER]);
  
  /* Speed */
  SDI_F->wheelSpeed.vr = DECI_IN_TO_M(State[STATE_VEL_RIGHT]);
  SDI_F->wheelSpeed.vl = DECI_IN_TO_M(State[STATE_VEL_LEFT]);

  posId->v = (SDI_F->wheelSpeed.vr + SDI_F->wheelSpeed.vl)/2.0;
  posId->w = (SDI_F->wheelSpeed.vr - SDI_F->wheelSpeed.vl)/SCOUT_DIAMETER;
   
  ds = posId->v * scoutExecTaskPeriod;
  dt = posId->w * scoutExecTaskPeriod;

  /* Calculer la position du point en avant (ou en arriere) */
  sens = SDI_F->ref.backFlag == SCOUT_BACKWARD_MOTION ? -1 : 1;
  posId->x = posId->xC + sens * distPoint * cos(posId->theta);
  posId->y = posId->yC + sens * distPoint * sin(posId->theta);

  /* Variance sur la position odo */
  odoProba(posId, ds, dt, SCOUT_DIAMETER,
	   SCOUT_VAR_COEFF,
	   &SDI_F->odo.odoError);

  /* C'est tout */
  return (OK);
}


/*------------------------------------------------------------------------
 * SonarsOn
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* scoutSonarsOnExec  -  codel EXEC de SonarsOn
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT scoutSonarsOnExec (int *bilan)
{
  conf_sn(SDI_F->sonarsFiring.fireRate, SDI_F->sonarsFiring.firedSonars);
  return ETHER;
}

/*------------------------------------------------------------------------
 * SonarsOff
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* scoutSonarsOffExec  -  codel EXEC de SonarsOff
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT scoutSonarsOffExec (int *bilan)
{
  int sn_order[16];
  sn_order[0] = 255;
  conf_sn(0, sn_order);
  return ETHER;
}

/*------------------------------------------------------------------------
 * scoutGetBatteriesLevel
 *
 */
static void scoutGetBatteriesLevel(int printFlag)
{
  int motorStatus;
  int b0, b1;
  
  SDI_F->batteriesLevel.levelCpu = (double) voltCpuGet();
  SDI_F->batteriesLevel.levelMotor = (double) voltMotorGet();
  motorStatus = State[STATE_MOTOR_STATUS];
  b0 = motorStatus & 4;
  b1 = motorStatus & 8;
  SDI_F->batteriesLevel.charge = 
    b0 ? SCOUT_BATTERY_HIGHT :
    b1 ? SCOUT_BATTERY_MEDIUM : 
    SCOUT_BATTERY_LOW;

  /* XXXXXXX */
  SDI_F->batteriesLevel.charge = SCOUT_BATTERY_MEDIUM;

  if (printFlag) {
    printf ("CPU %f V MOTOR %f V \n"
	    "status : motor right %d motor left %d \n"
	    "         batteries %s\n",
	    SDI_F->batteriesLevel.levelCpu, 
	    SDI_F->batteriesLevel.levelMotor, 
	    motorStatus & 1, motorStatus & 2,
	    SDI_F->batteriesLevel.charge == SCOUT_BATTERY_HIGHT ? "hight" : 
	    SDI_F->batteriesLevel.charge == SCOUT_BATTERY_MEDIUM ? "medium" : 
	    "low");
  }
}
